<?php

declare(strict_types=1);

namespace Cin\KBAFinder5;

use Exception;
use stdClass;
use Throwable;

final class TecDocCrawl
{
  private $providerid;

  private $apikey;

  private $articlecountry;

  private $serviceUrl;

  private $store;

  private $storeFile;

  private $limit = 250;

  public function __construct($providerid, $articlecountry, $apikey = null)
  {
    $this->storeFile      = __DIR__.'/tecdoccrawlstore.json';
    $this->serviceUrl     = 'https://webservice.tecalliance.services/pegasus-3-0/services/TecdocToCatDLB.jsonEndpoint';
    $this->providerid     = $providerid;
    $this->apikey         = $apikey;
    $this->articlecountry = $articlecountry;
  }

  public function crawl()
  {
    ini_set('max_execution_time', '0');

    $this->restore();

    if (0 === $this->store['currentPhase'] && $this->store['last_run'] > time() - (10 * 24 * 60 * 60)) {
      return false;
    }
    if (0 === $this->store['currentPhase']) {
      $this->store['currentPhase'] = 1;
    }

    switch ($this->store['currentPhase']) {
      case 1:
        $this->store['data'] = $this->getManufacturers();
        foreach ($this->store['data'] as &$manufacturer) {
          $manufacturer['models'] = $this->getModelSeries($manufacturer['manuId'])['array'];
          $manufacturer['done']   = false;
        }
        $this->store['currentPhase'] = 2;

        break;
      case 2:
        $count = 0;
        while ($count < $this->limit) {
          foreach ($this->store['data'] as &$manufacturer) {
            if ($manufacturer['done']) {
              continue;
            }
            foreach ($manufacturer['models'] as &$model) {
              if (isset($model['types'])) {
                continue;
              }
              $model['types'] = $this->getVehicleIdsByCriteria(
                $manufacturer['manuId'],
                $model['modelId']
              )['array'];
              ++$count;
              if ($count >= $this->limit) {
                break;
              }
            }
            $remaining = \count(array_filter($manufacturer['models'], static function ($model) {
              return ! isset($model['types']);
            }));
            if (0 === $remaining) {
              $manufacturer['done'] = true;
            }
            if ($count >= $this->limit) {
              break;
            }
          }
          ++$count;
        }
        $phaseFinished = 0 === \count(array_filter($this->store['data'], static function ($manufacturer) {
          return ! $manufacturer['done'];
        }));
        if ($phaseFinished) {
          $this->store['data']         = $this->flatten($this->store['data']);
          $this->store['currentPhase'] = 3;
        }

        break;
      case 3:
        $done = $this->getVehicleDetails();
        if ($done) {
          $this->store['currentPhase'] = 0;
          $this->store();

          return $this->store['data'];
        }

        break;
    }
    $this->store();

    return false;
  }

  public static function reset(): void
  {
    $fileName = __DIR__.'/tecdoccrawlstore.json';
    if (file_exists($fileName)) {
      unlink($fileName);
    }
  }

  private function getManufacturers()
  {
    $payload                                      = new stdClass();
    $payload->getManufacturers                    = new stdClass();
    $payload->getManufacturers->linkingTargetType = 'PO';

    return $this->callTecDoc($payload)['array'];
  }

  private function getModelSeries($manuid)
  {
    $payload                                    = new stdClass();
    $payload->getModelSeries                    = new stdClass();
    $payload->getModelSeries->manuId            = $manuid;
    $payload->getModelSeries->linkingTargetType = 'PO';

    return $this->callTecDoc($payload);
  }

  private function getVehicleIdsByCriteria($manuid, $modelid)
  {
    $payload                                                 = new stdClass();
    $payload->getVehicleIdsByCriteria                        = new stdClass();
    $payload->getVehicleIdsByCriteria->countriesCarSelection = $this->articlecountry;
    $payload->getVehicleIdsByCriteria->manuId                = $manuid;
    $payload->getVehicleIdsByCriteria->modId                 = $modelid;
    $payload->getVehicleIdsByCriteria->carType               = 'PO';

    return $this->callTecDoc($payload);
  }

  private function flatten($manufacturers)
  {
    $list = [];
    foreach ($manufacturers as $manufacturer) {
      foreach ($manufacturer['models'] as $model) {
        foreach ($model['types'] as $type) {
          $lineItem = [
            $manufacturer['manuName'],
            $model['modelname'],
            $type['carName'],
            $type['carId'],
          ];
          $list[$type['carId']] = $lineItem;
        }
      }
    }

    return $list;
  }

  private function getVehicleDetails()
  {
    $chunks = array_chunk(array_filter($this->store['data'], static function ($type) {
      return 4 === \count($type);
    }), 25, true);
    if (0 === \count($chunks)) {
      return true;
    }
    $max = min($this->limit, \count($chunks));
    for ($i = 0; $i < $max; ++$i) {
      $chunk                                            = $chunks[$i];
      $payload                                          = new stdClass();
      $payload->getVehicleByIds3                        = new stdClass();
      $payload->getVehicleByIds3->articleCountry        = $this->articlecountry;
      $payload->getVehicleByIds3->countriesCarSelection = $this->articlecountry;
      $payload->getVehicleByIds3->kbaData               = true;
      $payload->getVehicleByIds3->carIds                = new stdClass();
      $payload->getVehicleByIds3->carIds->array         = array_values(array_map(static function ($car) {
        return $car[3];
      }, $chunk));

      $details = $this->callTecDoc($payload)['array'];
      foreach ($details as $detail) {
        $this->store['data'][$detail['carId']][3] = $detail['vehicleDetails']['ccmTech'] ?? '';
        $this->store['data'][$detail['carId']][]  = $detail['vehicleDetails']['powerHpTo'] ?? '';
        $this->store['data'][$detail['carId']][]  = $detail['vehicleDetails']['powerKwTo'] ?? '';
        $this->store['data'][$detail['carId']][]  = $detail['vehicleDetails']['yearOfConstrFrom'] ?? '';
        $this->store['data'][$detail['carId']][]  = $detail['vehicleDetails']['yearOfConstrTo'] ?? '';
        $this->store['data'][$detail['carId']][]  = isset($detail['kbaNumbers']) ? array_map(static function ($kba) {
          return $kba['kbaNo'];
        }, $detail['kbaNumbers']['array']) : [];
        $this->store['data'][$detail['carId']][] = $detail['vehicleDetails']['typeNumber'];
      }
    }

    return false;
  }

  private function callTecDoc($payload)
  {
    reset($payload);
    $action                       = key($payload);
    $payload->{$action}->country  = $this->articlecountry;
    $payload->{$action}->lang     = 'DE';
    $payload->{$action}->provider = $this->providerid;

    try {
      $ch = curl_init($this->serviceUrl);
      curl_setopt($ch, \CURLOPT_RETURNTRANSFER, true);
      curl_setopt($ch, \CURLOPT_CUSTOMREQUEST, 'POST');
      curl_setopt($ch, \CURLOPT_POSTFIELDS, json_encode($payload));
      if (null !== $this->apikey && '' !== $this->apikey) {
        curl_setopt($ch, \CURLOPT_HTTPHEADER, ['x-api-key: '.$this->apikey]);
      }
      $result = curl_exec($ch);
      curl_close($ch);
      if (false === $result) {
        throw new Exception('Error Processing Licence Request', 1);
      }
      $response = json_decode($result, false);
      if (false === $response || null === $response) {
        throw new Exception('Error Decoding Licence Response', 2);
      }
    } catch (Throwable $th) {
      out($th->message."\n".$th->curlErrorString);
      exit(-2);
    } finally {
      if (200 !== $response->body['status']) {
        out(json_encode($response->body, \JSON_PRETTY_PRINT));
      }

      return $response->body['data'];
    }
  }

  private function restore(): void
  {
    if (file_exists($this->storeFile)) {
      $this->store = json_decode(file_get_contents($this->storeFile), true);
    } elseif ( ! isset($this->store['last_run'])) {
      $this->store                 = [];
      $this->store['currentPhase'] = 0;
      $this->store['last_run']     = 0;
      $this->store['data']         = [];
    }
  }

  private function store(): void
  {
    $this->store['last_run'] = time();
    file_put_contents($this->storeFile, json_encode($this->store, \JSON_UNESCAPED_SLASHES | \JSON_UNESCAPED_UNICODE | \JSON_PRETTY_PRINT));
  }
}
